@@ -2,6 +2,7 @@
 
 #from __future__ import division, print_function
 import warnings
+from copy import copy
 import numpy as np
 from astropy.coordinates import SkyCoord, Angle
 import astropy.units as u
@@ -13,6 +14,13 @@ import lmfit as lm
 from lmfit import Minimizer, Parameters, report_fit
 import PRF
 import sys
+import pygmmis
+import pkg_resources
+
+#Read in extrinsic error model from fitting thousands of binary systems
+_error_fname = pkg_resources.resource_filename(__name__, "error_model.npz")
+error_ext = pygmmis.GMM(K=2, D=2)
+error_ext.load(_error_fname)
 
 class PCA:
     """Class designed to give users access to analysis functions.
@@ -26,7 +34,7 @@ class PCA:
         Units of the frequencies in frequencies list.
     principal_components: int
         Number of components used in PCA for TPF lightcurve.
-    aperture: 2D Boolean array
+    aperture: 2D Boolean array, 'auto'
         If not specified user the TPF.pipeline_mask will be used, if a user specified aperture is used it must be the same shape as the TPF.
     Returns
     ----------
@@ -61,9 +69,37 @@ class PCA:
             
         if self.aperture is None:
             self.aperture = targetpixelfile.pipeline_mask
-            if targetpixelfile.pipeline_mask.any() == False:
+            if (targetpixelfile.pipeline_mask.any() == False):
                 #will add a flag here if no aperture
-                self.aperture = self.tpf.create_threshold_mask()
+                def frequency_aperture(tpf,frequencies,frequnits = 1/u.d):
+                    heat = np.empty((tpf.shape[1],tpf.shape[2]))
+                    heat[:]=np.nan
+                    #Iterating through columns of pixels
+
+                    for i in np.arange(0,tpf.shape[1]):
+
+                        #Iterating through rows of pixels
+                        for j in np.arange(0,tpf.shape[2]):
+
+
+                            #Making an empty 2-d array
+                            mask = np.zeros((tpf.shape[1],tpf.shape[2]), dtype=bool)
+
+                            #Iterating to isolate pixel by pixel to get light curves
+                            mask[i][j] = True
+
+                            #Getting the light curve for a pixel and excluding any flagged data
+                            lightcurve = tpf.to_lightcurve(aperture_mask=mask)
+                            lightcurve = lightcurve[np.isfinite(lightcurve['flux']*lightcurve['flux_err'])]
+                            lightcurve = lightcurve[np.where(lightcurve[np.isfinite(lightcurve['flux']*lightcurve['flux_err'])].quality==0)]
+                            pg = lightcurve.to_periodogram(frequency = frequencies,freq_unit = frequnits,ls_method='slow')
+
+                            heat[i][j] = np.sum(pg.power.value**2)**(1/2)
+                    return heat>np.mean(heat)+2*np.std(heat)
+                self.aperture = frequency_aperture(tpf=self.tpf,frequencies = frequencies, frequnits = frequnit)
+                    
+        if aperture =='auto':
+            self.aperture = frequency_aperture(tpf=self.tpf,frequencies = frequencies, frequnits = frequnit)
     
         
         # Make a design matrix and pass it to a linear regression corrector
@@ -139,9 +175,9 @@ class Localize:
         List of frequencies desired to localize the source location for.
     frequnit: astropy.units.unit
         Units of the frequencies in frequencies list.
-    principal_components: int
-        Number of components used in PCA for TPF lightcurve.
-    aperture: 2D Boolean array
+    principal_components: int, str
+        Number of components used in PCA for TPF lightcurve, or 'auto' for automatic determination.
+    aperture: 2D Boolean array, or 'auto'
         If not specified user the TPF.pipeline_mask will be used, if a user specified aperture is used it must be the same shape as the TPF.
 
     Returns
@@ -159,7 +195,7 @@ class Localize:
     self.frequency_list
         List of frequencies used.
     self.principal_components
-        Number of principal components removed.
+        Number of principal components removed, or 'auto'
     self.initial_phases
         Initial phases fit for the frequencies.
     self.final_phases
@@ -171,19 +207,23 @@ class Localize:
     self.timeserieslength
     self.gaiadata
     self.location
-        Best fit source location.
+        Best fit source location in pixels.
+    self.location_skycoord
+        Best fit source location in RA and DEC.
     self.heatmap
         2D array of composite heatmap for all frequencies.
     self.starfit
         Gaia sources and their distances from the fitted location of the source.
     self.result
-        Result paramters of the fit. Use report_fit(self.report) to view.
+        Result parameters of the fit. Use report_fit(self.report) to view.
+    self.maxsignal_aperture
+        Aperture mask for the pixel with the greatest SNR
     
     
     """
     
     def __init__(self, targetpixelfile, gaia=True, magnitude_limit=18, 
-                 frequencies=[], frequnit=u.uHz, principal_components = 5, 
+                 frequencies=[], frequnit=u.uHz, principal_components = 'auto', 
                  aperture=None, method = 'PRF', sigma=None, **kwargs):
         
         self.tpf = targetpixelfile
@@ -192,12 +232,47 @@ class Localize:
         self.aperture = aperture
         self.frequency_list = np.asarray((frequencies*frequnit).to(1/u.d))
         self.principal_components = principal_components
+            
+        if self.aperture is None:
+            self.aperture = targetpixelfile.pipeline_mask
+            if (targetpixelfile.pipeline_mask.any() == False):
+                #will add a flag here if no aperture
+                def frequency_aperture(tpf,frequencies,frequnits = 1/u.d):
+                    heat = np.empty((tpf.shape[1],tpf.shape[2]))
+                    heat[:]=np.nan
+                    #Iterating through columns of pixels
+
+                    for i in np.arange(0,tpf.shape[1]):
+
+                        #Iterating through rows of pixels
+                        for j in np.arange(0,tpf.shape[2]):
+
+
+                            #Making an empty 2-d array
+                            mask = np.zeros((tpf.shape[1],tpf.shape[2]), dtype=bool)
+
+                            #Iterating to isolate pixel by pixel to get light curves
+                            mask[i][j] = True
+
+                            #Getting the light curve for a pixel and excluding any flagged data
+                            lightcurve = tpf.to_lightcurve(aperture_mask=mask)
+                            lightcurve = lightcurve[np.isfinite(lightcurve['flux']*lightcurve['flux_err'])]
+                            lightcurve = lightcurve[np.where(lightcurve[np.isfinite(lightcurve['flux']*lightcurve['flux_err'])].quality==0)]
+                            pg = lightcurve.to_periodogram(frequency = frequencies,freq_unit = frequnits,ls_method='slow')
+
+                            heat[i][j] = np.sum(pg.power.value**2)**(1/2)
+                    return heat>np.mean(heat)+2*np.std(heat)
+                self.aperture = frequency_aperture(tpf=self.tpf,frequencies = frequencies, frequnits = frequnit)
+                    
+        if aperture =='auto':
+            self.aperture = frequency_aperture(tpf=self.tpf,frequencies = frequencies, frequnits = frequnit)
+            
         if principal_components == 'auto':
             self.principal_components = 5
             self.raw_lc1 = self.tpf.to_lightcurve(aperture_mask=self.aperture)
             self.quality_mask = [np.isfinite(self.raw_lc1['flux']*self.raw_lc1['flux_err'])]
             if self.principal_components !=0:
-                self.dm = lk.DesignMatrix(self.tpf.flux[:,~self.tpf.pipeline_mask][np.isfinite(self.raw_lc1['flux']*self.raw_lc1['flux_err']),:][np.where(self.raw_lc1[self.quality_mask[0]].quality ==0)[0],:], name='regressors').pca(self.principal_components)
+                self.dm = lk.DesignMatrix(self.tpf.flux[:,~self.aperture][np.isfinite(self.raw_lc1['flux']*self.raw_lc1['flux_err']),:][np.where(self.raw_lc1[self.quality_mask[0]].quality ==0)[0],:], name='regressors').pca(self.principal_components)
                 self.raw_lc = self.raw_lc1[self.quality_mask[0]]
                 self.raw_lc = self.raw_lc[np.where(self.raw_lc1[self.quality_mask[0]].quality ==0)]
 
@@ -238,7 +313,7 @@ class Localize:
         self.raw_lc1 = self.tpf.to_lightcurve(aperture_mask=self.aperture)
         self.quality_mask = [np.isfinite(self.raw_lc1['flux']*self.raw_lc1['flux_err'])]
         if self.principal_components !=0:
-            self.dm = lk.DesignMatrix(self.tpf.flux[:,~self.tpf.pipeline_mask][np.isfinite(self.raw_lc1['flux']*self.raw_lc1['flux_err']),:][np.where(self.raw_lc1[self.quality_mask[0]].quality ==0)[0],:], name='regressors').pca(self.principal_components)
+            self.dm = lk.DesignMatrix(self.tpf.flux[:,~self.aperture][np.isfinite(self.raw_lc1['flux']*self.raw_lc1['flux_err']),:][np.where(self.raw_lc1[self.quality_mask[0]].quality ==0)[0],:], name='regressors').pca(self.principal_components)
             self.raw_lc = self.raw_lc1[self.quality_mask[0]]
             self.raw_lc = self.raw_lc[np.where(self.raw_lc1[self.quality_mask[0]].quality ==0)]
         
@@ -446,6 +521,7 @@ class Localize:
         heats_error = np.asarray(heats_error)
         self.heats = heats.T
         self.heats_error = heats_error.T
+        
         self.timeserieslength = (self.tpf.time.max()-self.tpf.time.min()).value
         self.gaiadata = None
         
@@ -500,6 +576,7 @@ class Localize:
 
 
             self.gaiadata = source
+        
         class frequency_heatmap:
 
             def __init__(self,tpf,heats,heats_error,frequencies,gaia_data,method):
@@ -524,8 +601,8 @@ class Localize:
                     #Residuals to minimize relative to the error bars
                     def residual(params, amp, amperr, prf):
 
-                        x = params['x']
-                        y = params['y']
+                        x = params['column']
+                        y = params['row']
 
                         res = []
                         localprf = self.prf.locate(x, y, (self.size[0],self.size[1]))
@@ -549,8 +626,8 @@ class Localize:
                     params = Parameters()
                     for i in np.arange(len(frequencies)):#
                         params.add('height{0:d}'.format(i), value=np.max(self.heat_stamp[i]))
-                    params.add('x', value=c[1][0])#c[0]) 
-                    params.add('y', value=c[0][0])#c[1])
+                    params.add('column', value=c[1][0])#c[0]) 
+                    params.add('row', value=c[0][0])#c[1])
                     #params.add('sigma', value=1)
                 else:
                     if sigma is None:
@@ -562,8 +639,8 @@ class Localize:
                     #Residuals to minimize relative to the error bars
                     def residual(params, amp, amperr, prf):
 
-                        x = params['x']
-                        y = params['y']
+                        x = params['column']
+                        y = params['row']
                         #sigma = params['sigma'] #KJB removed
                         
                         res = []
@@ -587,8 +664,8 @@ class Localize:
                     params = Parameters()
                     for i in np.arange(len(frequencies)):#
                         params.add('height{0:d}'.format(i), value=np.max(self.heat_stamp[i]))
-                    params.add('x', value=c[1][0])#c[0]) 
-                    params.add('y', value=c[0][0])#c[1])
+                    params.add('column', value=c[1][0])#c[0]) 
+                    params.add('row', value=c[0][0])#c[1])
                     #params.add('sigma', value=1)        # KJB: not fitted
                     
                     
@@ -596,11 +673,11 @@ class Localize:
                 minner = Minimizer(residual, params, fcn_args=(self.heat_stamp, self.heatmap_error, self.prf))
                 self.result = minner.minimize()
                 fit = self.result.params.valuesdict()
-                self.x = fit['x']
-                self.y = fit['y']
+                self.x = fit['column']
+                self.y = fit['row']
 
                 
-            def star_list(self):
+            def star_list(self, logL):
                 gaia_data = self.gaiadata
                 no_gaia_data_message = ValueError('No gaia data initialized in PixelMapPeriodogram class')
                 if gaia_data ==None :
@@ -608,30 +685,44 @@ class Localize:
 
                 else:
                     distances = np.square(self.x-gaia_data['x'])+np.square(self.y-gaia_data['y'])
-                    closest_star_mask = np.where(np.square(self.x-gaia_data['x'])+np.square(self.y-gaia_data['y'])==(np.square(self.x-gaia_data['x'])+np.square(self.y-gaia_data['y'])).min())
+                    #closest_star_mask = np.where(np.square(self.x-gaia_data['x'])+np.square(self.y-gaia_data['y'])==(np.square(self.x-gaia_data['x'])+np.square(self.y-gaia_data['y'])).min())
                     stars = dict(ra = np.asarray(gaia_data['ra']),
                                  dec = np.asarray(gaia_data['dec']),
                                  source = np.asarray(gaia_data['source']),
                                  x = np.asarray(gaia_data['x']),
                                  y = np.asarray(gaia_data['y']),
+                                 Gmag = np.asarray(gaia_data['Gmag']),
                                  distance = distances)
-                                 #include a probability metric here eventually 
-                        
+                    #compute likelihoods of gaia sources
+                    L = 10**logL(np.vstack((gaia_data["x"],gaia_data["y"])).T) #likelihoods
+                    L /= np.sum(L) #normalized
+                    stars["likelihood"] = L
                     starlist = pd.DataFrame.from_dict(stars)
-                    self.stars = starlist.sort_values(by=[r'distance'])
+                    self.stars = starlist.sort_values(by=[r'likelihood'],ascending = False)
                     
+        
         fh = frequency_heatmap(self.tpf,self.heats,self.heats_error,self.frequency_list,self.gaiadata,self.method) 
         fh.location()
         self.location = [fh.x,fh.y]
+        self.location_skycoord = self.tpf.wcs.all_pix2world([[fh.x,fh.y]], 0)[0]
         self.heatmap = self.heats.sum(axis=0).reshape(self.aperture.shape[0],self.aperture.shape[1]) / np.sqrt((self.heats_error**2).sum(axis=0)).reshape(self.aperture.shape[0],self.aperture.shape[1])
+        self.maxsignal_aperture = self.heatmap == self.heatmap.max()
         self.result = fh.result
-        if (self.gaiadata !=None):
-            fh.star_list()
+        
+        #Combined error model
+        error_model = copy(error_ext) #extrinsic error
+        error_model.covar += self.result.covar[-2:,-2:] #intrinsic error
+        error_model.mean += self.location #Locate relative to best-fit position
+        self.logL = error_model.logL
+        
+        if (self.gaiadata is not None):
+            fh.star_list(self.logL)
             self.starfit= fh.stars.reset_index()
 
     
     def info(self):
         plt.imshow(self.heatmap,origin='lower')
+        plt.title('SNR')
         #plot the location
         if (self.gaiadata != None):
             plt.scatter(self.gaiadata['x'],self.gaiadata['y'],s=self.gaiadata['size']*5,c='white',alpha=.6)
@@ -641,7 +732,10 @@ class Localize:
         plt.ylim(-.5,self.aperture.shape[0]-1+.5)
         report_fit(self.result)
         if (np.asarray([self.result.params['height{0:d}'.format(j)].stderr for j in np.arange(len(self.frequency_list))]) / np.asarray([self.result.params['height{0:d}'.format(j)].value for j in np.arange(len(self.frequency_list))])>.2).any():
+            #possibly reword this
             warnings.warn('Frequencies used may not all belong to the same source and provided fit could be unreliable')
+        if ((self.location[0]<0) and (self.location[0]>tpf.shape[1])) or ((self.location[1]<0) and (self.location[1]>tpf.shape[2])):
+            warnings.warn('Source fit to a location outside the TPF, refitting using a TPF centered around source is recommended')
     
     def pca(self):
         if self.principal_components==0:
